/*
 * ----------------------------------------------------------------------------
 
 */

#include "defines.h"

#include <avr/interrupt.h>
#include <ctype.h>
#include <stdint.h>
#include <stdio.h>
#include <math.h>
#include <inttypes.h>
#include <stdlib.h>
#include <avr/eeprom.h>

#include <avr/io.h>
#include <avr/pgmspace.h>
#define SYSCLK 16000000UL			// Prozessortakt fr delay festlegen
#include <util/delay.h>

#include "lcd.h"
#include "uart.h"


volatile const char SW_UP = 0;			// Taster 'UP'
volatile const char SW_DOWN = 1;		// Taster 'DOWN'
volatile const char SW_ENTER = 2;		// Taster 'ENTER'
volatile const char SW_EXIT = 3;		// Taster 'EXIT'
volatile const char P_PRIM = 7;			// Analogeingang Primrdruck
volatile const char P_SEK = 6;			// Analogeingang Sekundrdruck
volatile const char SSR_OUT = 4;		// Solid-State-Relais-Ausgang
volatile const char MV1 = 5;			// Magnetventil 1 - Ausgang
volatile const char MV2 = 4;			// Magnetventil 2 - Ausgang

volatile const char ST_EDIT = 0;		// Merkervariable 0 => 'EDIT'-Modus
volatile const char ST_MESS = 1;		// Merkervariable 1 => 'AUSGABE'-Modus auf Display
volatile const char ST_LZ = 2;			// Merkervariable 2 => 'Laufzeit angesprochen'
volatile const char ST_MUF = 3;			// Merkervariable 3 => 'MU - Fehler'
volatile const char ST_ICP = 4;			// Merkervariable 4 => ICP gestartet
volatile const char ST_w = 5;			// Merkervariable 5 => Men - Sollwerteingabe ausgewhlt
volatile const char ST_Kp = 6;			// Merkervariable 5 => Men - Kp-Eingabe ausgewhlt
volatile const char ST_I = 7;			// Merkervariable 5 => Men - I-Eingabe ausgewhlt
volatile const char ST_D = 8;			// Merkervariable 5 => Men - D-Eingabe ausgewhlt
volatile const char ST_AS = 9;			// Merkervariable 5 => Men - D-Eingabe ausgewhlt

volatile double ADC_temp = 0;
volatile double Uneu = 0;
volatile double Strom = 0;
double Sek_STemp = 2.5;
volatile double Sek_X = 0;				// Regelabweichung Sekundr
volatile unsigned int Y_Zul = 0;			// Stellgre Zuluft
volatile unsigned int Y_Abl = 0;			// Stellgre Abluft
volatile double ASW = 0.1;				// Ansprechschwelle
unsigned long Verst = 20480;					// Verstrkung
volatile unsigned int Status = 0;
volatile float Dec_St = 0.1;
volatile unsigned char D_St = 0;
volatile unsigned char RG_CNT = 1;
volatile unsigned char DISP_CNT = 1;
volatile unsigned int L_ZEIT = 0;			// Laufzeit (max. 15 min.)
volatile unsigned int R_ZEIT = 0;			// Ruhezeit (15 min.)
volatile unsigned char DST_cnt = 0;			// Dezimalstellenzhler Men
// PT100-Berechnung ----------------------------------------------------
double const A = 3.9083E-3;
double const B = -5.775E-7;
double const C = -4.183E-12;
double const R0 = 100;		// Widerstand bei 0C
double T = 0;			// Temperatur bei 0C

//--- Regelparameter ---------------------------------------------------
double Strombereich = 0.01953125;			// 20mA / 1024
double Sek_Soll = 70;						// Sollwert
volatile double Sek_Ist = 0;				// Istwert Sekundr
double Pr_Soll = 0;							// Sollwert Primr
volatile double Pr_Ist = 0;					// Istwert Primr
volatile double e = 0;						// Regelabweichung
volatile double w = 0;						// Sollwert
volatile double x = 0;						// Istwert
volatile double y = 0;						// Stellgre
volatile double ealt = 0;					// vorherige Regelabweichung
volatile double esum = 0;					// Summe Regelabweichung
volatile double Ta = .02;					// Abtastzeit 20ms
double Kp = 2;								// Verstrkung (.04) 
double I = 1;								// Integralanteil (.01)
double D = 0;								// Differenzanteil (.01)
double AS = 1;								// Ansprechschwelle
double Integral = 2.4;						// Berechneter Integralanteil (I * Ta)
volatile unsigned int starttime = 0;		// Zhler vom ICR
volatile unsigned int stoptime = 0;			// vorheriger Zhler vom ICR
volatile double Frequenz = 0;				// vorheriger Zhler vom ICR
volatile unsigned int ICP_OVF = 0;			// Input-capture-Overflow-Zhler
unsigned char i = 0;
double Dummy = 0;

/*** Hardwareinitialisierung ***************************************************************/

static void ioinit(void)
{
  uart_init();
  lcd_init();
  
 // PORTA Datenrichtungen & Pullup

  DDRA &= ~_BV(SW_UP)&~_BV(SW_DOWN)&~_BV(SW_ENTER)&~_BV(SW_EXIT)&~_BV(P_PRIM)&~_BV(P_SEK);
  PORTA = _BV(SW_UP)|_BV(SW_DOWN)|_BV(SW_ENTER)|_BV(SW_EXIT);
  
 // PORTB Datenrichtungen & Pullup
 
  PORTB &= ~_BV(SSR_OUT);
  DDRB |= (1<<SSR_OUT);
 
 // PORTC Datenrichtungen & Pullup
 
	// Im Displaymodul definiert
 
 // PORTD Datenrichtungen & Pullup
  
  DDRD &= ~_BV(PD6);
  PORTD &= ~_BV(MV1)&~_BV(MV2);
  //PORTD |= _BV(PD6);
  DDRD = _BV(MV1)|_BV(MV2);
  PORTD |= _BV(MV1)|_BV(MV2);
  
 // Einstellungen fr AD-Wandler	
	
	ADCSRA = 0b10000110;				// '110' 115.2kHz bei 7.3728 MHz (/64)
	ADMUX = 0b10000000;					// ADC-Kanal 0, ADC enabled
	
 // TIMER0-Einstellungen

	TCCR0 = _BV(CS00)|_BV(CS02);		// Timer0 (CS00 CS02 => /1024)
	
 // TIMER1-Einstellungen

	TCCR1A = _BV(COM1A1)|_BV(COM1B1)|_BV(WGM10)|_BV(WGM11); 	// 3,7kHz, non-inverted Phase & Freq. correct PWM
	TCCR1B = _BV(CS10)|_BV(ICNC1)|_BV(ICES1); 					// prescaler /1
	OCR1A = 0;
	OCR1B = 0;
	
	TIFR = _BV(1<<TOV0)|_BV(ICF1);		// Flag lschen	
	
	#include <PT100.c>				// Temperaturberechnung einbinden
	Dummy = PT100_Temp(280.90);		// Widerstand bergeben, Temperatur zurckbekommen
	
}

/**************************************************************************************************/
/*** Zuweisung der IN- Out - Streams **************************************************************/

FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);
FILE lcd_str = FDEV_SETUP_STREAM(lcd_putchar, NULL, _FDEV_SETUP_WRITE);

/**************************************************************************************************/

static void delay_1s(void)			// Eine Sekunde Pause
{
  uint8_t i;

  for (i = 0; i < 100; i++)
    _delay_ms(10);
}

/**** ADC *****************************************************************************************/

double ADC_conversion(unsigned char mux)			// Spannung messen (32 Messswerte-Mittelwert)
{	
	ADMUX = mux;						// Multiplexer einstellen 
	unsigned int ADC_temp;
    Uneu = 0;
    
    for(unsigned char i=0;i<=31;i++)   // 32 Messungen fr bessere Genauigkeit 
		{
			ADCSRA |= (1<<ADSC);       // 'single conversion' einleiten
			while(!(ADCSRA & 0x10));  	// auf ende der Wandlung warten, ADIF flag '1'
			ADC_temp = ADCL;           	// ADCL Register lesen
			ADC_temp += (ADCH << 8);   // ADCH Register lesen      
			Uneu += ADC_temp;     		// Ergebnisse der 32 Messungen addieren
		}
 	Uneu = Uneu/32;						// Mittelwert der 32 Messungen der globalen Variable zuweisen
	
	Strom = Strombereich * Uneu;		// mA berechnen (0..20mA)
	
	return (Strom);
}

/**** Regler ***************************************************************************************/

double PID_Berechnung (double x, double w)
{		
	e = w - x;										// aktuelle Regelabweichung bestimmen
//--------------------------------------------------------------------------------------------------
	if ((e >= AS)||(e <= (AS*(-1))))				// Betrag der Differenz prfen
	{
//--------------------------------------------------------------------------------------------------
	if ((y < 1023)&&(y > 0))						// bei bersteuertem stellglied Integration einfrieren
		{											// (Anti-Windup)
			esum = esum + e;						// Summe der Regelabweichung aktualisieren	
		}
//--------------------------------------------------------------------------------------------------	 
		y = (Kp*e)+(I*Ta*esum)+(D*((e-ealt))/Ta);	// Reglergleichung 
													// (I-Anteil bei der Initialisierung mit Abtastrate verrechnet)
		ealt = e;									// Regelabweichung fr nchste Abtastung merken
	}
//--------------------------------------------------------------------------------------------------
	if (y > 1023)									// Stellgre auf 0..1023 begrenzen (10 bit PWM)
		{
			y = 1023;
		}
	if (y < 1)
		{
			y = 0;
		}
//--------------------------------------------------------------------------------------------------
	return y;
}

//--------------------------------------------------------------------------------------
/*** Regelung *************************************************************************/

void Regeln (void)
{
	OCR1B = PID_Berechnung (Frequenz, Sek_Soll);	// Regelparameter berechnen
}


//*************************************************************************************
//*************************************************************************************

int main(void)
{


  ioinit();
  UBRRL = 8;
  
  Status &= ~_BV(ST_LZ);

  stdout = stdin = &uart_str;
  stderr = &lcd_str;
 
  lcd_clr();
  lcd_setcrsr(0x80);
  fprintf(stderr, " PID-Regler V2.1");
  _delay_ms(1);
  lcd_setcrsr(0xc0);
  fprintf(stderr, "  THM 11/2007   ");
  delay_1s();
  delay_1s();
  delay_1s();

  eeprom_busy_wait();
  eeprom_read_block(&Sek_Soll,(uint8_t*)0,sizeof(double));
  eeprom_busy_wait();
  eeprom_read_block(&Kp,(uint8_t*)4,sizeof(double));
  eeprom_busy_wait();
  eeprom_read_block(&I,(uint8_t*)8,sizeof(double));
  eeprom_busy_wait();
  eeprom_read_block(&D,(uint8_t*)12,sizeof(double));

  if (!((Sek_Soll >= 0)&&(Sek_Soll <= 140)))
  {
	Sek_Soll = 50;
  }
  Integral = I*Ta;									// I-Anteil mit Abtastrate verrechnen
 
  lcd_clr();
  lcd_setcrsr(0x80);
  fprintf(stderr, "x: %7.1f Hz",ADC_conversion(7));
  _delay_ms(1);
  lcd_setcrsr(0xc0);
  fprintf(stderr, "y: %7.1f Hz  ",Sek_Ist);
  _delay_ms(1);

  TIMSK = _BV(TOIE0)|_BV(TICIE1)|_BV(TOIE1);		// Interrupts freigegeben
  
  SREG = _BV(SREG_I);								// Globalen Interrupt freigeben
  
  #include "Menue.c"
  
  for (;;)
    {	
/*--- Taste 'ENTER' ----------------------------------------------------------*/
	if (!(PINA &(1<<SW_ENTER)))
	{
	TCCR0 = 0;									// Timer0 gestoppt
	for (unsigned char i=0;i<=19;i++)
	{
		_delay_ms(10);
	}
	lcd_setcrsr(0x80);
	fprintf(stderr, " - Einstellung -");
		
		if (DST_cnt == 0)
		{
			Status |= _BV(ST_w);
			Menue();
			DST_cnt ++;
			goto Ende;
		}
		if (DST_cnt == 1)
		{
			Status |= _BV(ST_Kp);
			Menue();
			DST_cnt ++;
			goto Ende;
		}
		if (DST_cnt == 2)
		{
			Status |= _BV(ST_I);
			Menue();
			DST_cnt ++;
			goto Ende;
		}
		if (DST_cnt == 3)
		{
			Status |= _BV(ST_D);
			Menue();
			DST_cnt ++;
			goto Ende;
		}
		if (DST_cnt == 4)
		{
			Status |= _BV(ST_AS);
			Menue();
			DST_cnt = 0;
		}
	Ende:
	while (!(PINA &(1<<SW_ENTER)));
	}
	
/*--- Zyklischer Programmablauf ----------------------------------------------*/	
	if ((Status &(1<<ST_MESS)))
	{
		lcd_setcrsr(0x80);
		fprintf(stderr, "x:%6.1f w:%5.1f",Frequenz,Sek_Soll);
		_delay_ms(1);
		lcd_setcrsr(0xc0);
		fprintf(stderr, "y:%4.0f e:%5.1f   ",y,e);
		_delay_ms(1);
		lcd_setcrsr(0xd0);
		fprintf(stdout, "%4.1f %5.1f %5.1f %4.0f\n",Sek_Soll,Frequenz,e,y);// Daten zur Testaufzeichnung
		Status &= ~_BV(ST_MESS);
		
		if (!(Status &(1<<ST_MUF)))				// wenn MU-Fehler weg => weitermachen
		{
			lcd_setcrsr(0xc0);
			_delay_ms(1);
			Status &= ~_BV(ST_MUF);
		}
	}
    }
  return 0;
}


/**** Interrupt - Routinen **************************************************/

SIGNAL (SIG_OVERFLOW0)			// ca. 10ms - Timer
{
	TCNT0 = 178;				
	
	if (RG_CNT >= 2)			// 20ms - Takt
	{
		//Pr_Ist = ADC_conversion(7);
		//Sek_Ist = ADC_conversion(6);	// Beide Drcke messen
		Regeln();								
		RG_CNT = 1;
	}
	
	if (DISP_CNT >= 50)			// 500ms - Takt
	{
		if ((!(Status &(1<<ST_MUF))))
		{
			Status = _BV(ST_MESS);
		}
		DISP_CNT = 1;
	}

	RG_CNT++;
	DISP_CNT++;
}

//--- Input Capture --------------------------------------------------------------------------

SIGNAL (SIG_INPUT_CAPTURE1)		// Frequenzmessung
{
	
	if (!(Status &(1<<ST_ICP)))
	{
		starttime = ICR1;
		ICP_OVF = 0;
		Status |= _BV(ST_ICP);
		return;
	}
	else
	{
		stoptime = ICR1;
		Frequenz = (ICP_OVF * 65536) + stoptime - starttime;	// Zhlerstand ermitteln
		Frequenz = (512000000/Frequenz);						// SYSCLK*32 wegen PWM-Mode,
		Status &= ~_BV(ST_ICP);								// Frequenz berechnen
		TIFR |= _BV(ICF1);
	}
}

//--- Timer1 overflow -----------------------------------------------------------------------

SIGNAL (SIG_OVERFLOW1)			// ICP - Overflow-Zhler
{
	ICP_OVF++;
	if (ICP_OVF >5000)		// Timeout bei Sensordefekt erkennen
	{
		Frequenz = 0;
	}
}

